from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union

from pydantic import BaseModel, TypeAdapter

from thanosql._service import ThanoSQLService

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class QueryLog(BaseModel):
    query_id: Optional[str]
    statement_type: Optional[str]
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    query: str
    referer: str
    state: Optional[str]
    destination_table_name: Optional[str]
    destination_schema: Optional[str]
    error_result: Optional[str]
    created_at: Optional[datetime]
    records: Optional[list] = None


class QueryService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="query")

        self.log: QueryLogService = QueryLogService(self)
        self.template: QueryTemplateService = QueryTemplateService(self)

    def execute(
        self,
        query_type: str = "thanosql",
        query: Optional[str] = None,
        template_id: Optional[int] = None,
        template_name: Optional[str] = None,
        parameters: Optional[dict] = None,
        schema: Optional[str] = None,
        table_name: Optional[str] = None,
        overwrite: Optional[bool] = None,
        max_results: Optional[int] = None,
    ) -> Union[QueryLog, dict]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema,
            table_name=table_name,
            overwrite=overwrite,
            max_results=max_results,
        )
        payload = self._create_input_dict(
            query_type=query_type,
            query_string=query,
            template_id=template_id,
            template_name=template_name,
            parameters=parameters,
        )

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        if "query_id" in raw_response:
            query_log_adapter = TypeAdapter(QueryLog)
            parsed_response = query_log_adapter.validate_python(raw_response)
            return parsed_response

        return raw_response


class QueryLogService(ThanoSQLService):
    """Cannot exist without a parent QueryService"""

    def __init__(self, query: QueryService) -> None:
        super().__init__(client=query.client, tag="log")

        self.query: QueryService = query

    def list(
        self,
        search: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> dict:
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(
            search=search, offset=offset, limit=limit
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "query_logs" in raw_response:
            query_logs_adapter = TypeAdapter(List[QueryLog])
            parsed_response = {}
            parsed_response["query_logs"] = query_logs_adapter.validate_python(
                raw_response["query_logs"]
            )
            parsed_response["total"] = raw_response["total"]
            return parsed_response

        return raw_response


class QueryTemplate(BaseModel):
    id: Optional[int] = None
    name: str
    query: str
    parameters: Optional[List[str]] = []
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class QueryTemplateService(ThanoSQLService):
    """Cannot exist without a parent QueryService"""

    def __init__(self, query: QueryService) -> None:
        super().__init__(client=query.client, tag="template")

        self.query: QueryService = query

    def list(
        self,
        search: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> Union[List[QueryTemplate], dict]:
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(
            search=search, offset=offset, limit=limit, order_by=order_by
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "query_templates" in raw_response:
            query_templates_adapter = TypeAdapter(List[QueryTemplate])
            parsed_response = query_templates_adapter.validate_python(
                raw_response["query_templates"]
            )
            return parsed_response

        return raw_response

    def create(
        self,
        name: Optional[str] = None,
        query: Optional[str] = None,
        dry_run: Optional[bool] = None,
    ) -> Union[QueryTemplate, dict]:
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(dry_run=dry_run)
        payload = self._create_input_dict(name=name, query=query)

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        if "query_template" in raw_response:
            query_template_adapter = TypeAdapter(QueryTemplate)
            parsed_response = query_template_adapter.validate_python(
                raw_response["query_template"]
            )
            return parsed_response

        return raw_response

    def get(self, name: str) -> Union[QueryTemplate, dict]:
        path = f"/{self.query.tag}/{self.tag}/{name}"

        raw_response = self.client._request(method="get", path=path)

        if "query_template" in raw_response:
            query_template_adapter = TypeAdapter(QueryTemplate)
            parsed_response = query_template_adapter.validate_python(
                raw_response["query_template"]
            )
            return parsed_response

        return raw_response

    def update(
        self,
        current_name: str,
        new_name: Optional[str] = None,
        query: Optional[str] = None,
    ) -> Union[QueryTemplate, dict]:
        path = f"/{self.query.tag}/{self.tag}/{current_name}"
        payload = self._create_input_dict(name=new_name, query=query)

        raw_response = self.client._request(method="put", path=path, payload=payload)

        if "query_template" in raw_response:
            query_template_adapter = TypeAdapter(QueryTemplate)
            parsed_response = query_template_adapter.validate_python(
                raw_response["query_template"]
            )
            return parsed_response

        return raw_response

    def delete(self, name: str) -> dict:
        path = f"/{self.query.tag}/{self.tag}/{name}"

        return self.client._request(method="delete", path=path)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tests.faker import fake
from thanosql._error import (
    ThanoSQLAlreadyExistsError,
    ThanoSQLNotFoundError,
    ThanoSQLValueError,
)
from thanosql.resources import TableObject, TableTemplate

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL

latest_version = "2.0"


def test_create_table_template_invalid(
    client: ThanoSQL, empty_table_template: TableTemplate
):
    # check that we cannot create table template with too long of a name
    with pytest.raises(ThanoSQLValueError):
        client.table.template.create(
            name=fake.unique.pystr(min_chars=35, max_chars=40).lower(),
            table_template=TableObject(),
        )

    # check that we cannot create table template with invalid characters
    with pytest.raises(ThanoSQLValueError):
        client.table.template.create(name="!#%", table_template=TableObject())

    # creation of empty table template version 1.0 is done in conftest set-up
    name = empty_table_template.name

    # check that we cannot create table template with the same name and version
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.table.template.create(name=name, table_template=TableObject())

    # invalid version format should not work
    with pytest.raises(ThanoSQLValueError):
        client.table.template.create(
            name=name, table_template=TableObject(), version="2.0.0"
        )


def test_create_table_template_success(
    client: ThanoSQL, empty_table_template_name: str
):
    # creating the same template with different (valid) version should work
    res = client.table.template.create(
        name=empty_table_template_name,
        table_template=TableObject(),
        version=latest_version,
    )
    assert {"message", "table_template_name"} == set(res.keys())


def test_get_table_template_not_found(client: ThanoSQL):
    # random name that is yet to be used for creation should not work
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.get(name=fake.unique.pystr(8))


def test_get_table_template_not_found(client: ThanoSQL, empty_table_template_name: str):
    # random name that is yet to be used for creation should not work
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.get(name=fake.unique.pystr(8))

    # random version should not work
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.get(name=empty_table_template_name, version="3.0")


def test_get_table_template(client: ThanoSQL, empty_table_template_name: str):
    # latest should return the latest version only
    res = client.table.template.get(name=empty_table_template_name, version="latest")
    assert {"table_templates", "versions"} == set(res.keys())
    assert len(res["table_templates"]) == 1
    assert res["table_templates"][0].version == latest_version

    # only the specified version should be returned
    res = client.table.template.get(name=empty_table_template_name, version="1.0")
    assert {"table_templates", "versions"} == set(res.keys())
    assert len(res["table_templates"]) == 1
    assert res["table_templates"][0].version == "1.0"

    # without any version in the request, all table templates should be returned
    res = client.table.template.get(name=empty_table_template_name)
    assert {"table_templates", "versions"} == set(res.keys())
    assert len(res["table_templates"]) == 2
    assert len(res["versions"]) == 2
    assert res["versions"][0] == latest_version
    assert res["versions"][1] == "1.0"


def test_get_table_templates_default(client: ThanoSQL):
    res = client.table.template.list()
    assert isinstance(res, list)
    # at least the two templates that we made earlier
    assert len(res) >= 2


def test_delete_table_template_not_found(
    client: ThanoSQL, empty_table_template_name: str
):
    # check that we cannot delete table template with random name
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.delete(name=fake.unique.pystr(8))

    # random version should also not work
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.delete(name=empty_table_template_name, version="3.0")


def test_delete_table_template_success(
    client: ThanoSQL, empty_table_template_name: str
):
    # delete a specific version only
    res = client.table.template.delete(
        name=empty_table_template_name, version=latest_version
    )
    assert res["table_template_name"] == empty_table_template_name

    # now said version should not be found
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.get(
            name=empty_table_template_name, version=latest_version
        )

    # the other version should still exist
    # no need for comprehensive asserts as it is covered in previous tests already
    res = client.table.template.get(name=empty_table_template_name, version="1.0")
    assert res["versions"][0] == "1.0"

    # delete all versions
    res = client.table.template.delete(name=empty_table_template_name)
    assert res["table_template_name"] == empty_table_template_name

    # now any table template with this name should not exist
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.template.get(name=empty_table_template_name)



from __future__ import annotations

import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class ThanoSQLService:
    def __init__(self, client: ThanoSQL, tag: str = "") -> None:
        self.client: ThanoSQL = client
        self.tag: str = tag

    def _convert_obj_to_dict(self, obj: object) -> dict:
        try:
            model_dump = obj.model_dump_json(by_alias=True)
        except:
            model_dump = json.dumps(obj, default=lambda o: o.__dict__)
        return json.loads(model_dump)

    def _create_input_dict(self, **kwargs) -> dict:
        input_dict = {}
        for key, value in kwargs.items():
            if value is not None:
                if hasattr(value, "__dict__"):
                    value = self._convert_obj_to_dict(value)
                input_dict[key] = value

        return input_dict



from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING

import pytest

from tests.faker import fake
from thanosql._error import (
    ThanoSQLAlreadyExistsError,
    ThanoSQLNotFoundError,
    ThanoSQLValueError,
)
from thanosql.resources import (
    BaseColumn,
    BaseTable,
    Constraints,
    PrimaryKey,
    Table,
    TableObject,
)

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest

    from thanosql._client import ThanoSQL

test_table_name = f"test_table_{fake.unique.pystr(8).lower()}"
test_table_name_old = test_table_name + "_old"
test_table_name_excel = test_table_name + "_excel"


def test_get_table_not_found(client: ThanoSQL):
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.get(name=fake.unique.pystr(10))


@pytest.mark.parametrize("name", ["empty_table_name", "basic_table_name"])
def test_get_table_success(client: ThanoSQL, name: str, request: FixtureRequest):
    name = request.getfixturevalue(name)
    res = client.table.get(name=name)
    assert isinstance(res, Table)
    assert res.name == name


def test_create_table_invalid(client: ThanoSQL, empty_table_name: str):
    # creating a table that already exists should not be allowed
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.table.create(name=empty_table_name, schema="public", table=TableObject())

    # creating a table in a nonexistent schema should not be allowed
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.create(
            name=fake.unique.pystr(8), schema=fake.unique.pystr(8), table=TableObject()
        )

    # creating a table with invalid body should not be allowed
    with pytest.raises(ThanoSQLValueError):
        table_object = TableObject(
            columns=[BaseColumn(type="varchar", name="some_column")],
            constraints=Constraints(primary_key=PrimaryKey(columns=["another_column"])),
        )
        client.table.create(name=fake.unique.pystr(8), table=table_object)


def test_create_table_success(client: ThanoSQL, new_schema: str):
    # create a table in a non-"public" schema properly without fixture
    res = client.table.create(
        name=test_table_name_old, schema=new_schema, table=TableObject()
    )
    assert {"message", "table_name"} == set(res.keys())
    assert res["table_name"] == test_table_name_old

    # check that the table is indeed created
    res = client.table.get(name=test_table_name_old, schema=new_schema)
    assert res.name == test_table_name_old


def test_get_tables_default(client: ThanoSQL):
    res = client.table.list()
    assert isinstance(res, list)
    # at least the three tables that we made earlier
    assert len(res) >= 3


def test_get_tables_with_options(client: ThanoSQL, new_schema: str):
    res = client.table.list(schema=new_schema, verbose=True)
    assert isinstance(res, list)
    # we only have one table in this schema
    assert len(res) == 1
    assert isinstance(res[0], BaseTable)


def test_update_table(client: ThanoSQL, new_schema: str):
    table_object = BaseTable(
        name=test_table_name,
        schema="public",
        columns=[BaseColumn(type="integer", name="number")],
        constraints=Constraints(
            primary_key=PrimaryKey(
                name=f"pk_{fake.unique.pystr(8)}", columns=["number"]
            )
        ),
    )
    res = client.table.update(
        name=test_table_name_old, schema=new_schema, table=table_object
    )
    assert {"message", "table_name"} == set(res.keys())
    assert res["table_name"] == test_table_name_old

    # make sure the old table is no longer retrievable
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.get(name=test_table_name_old, schema=new_schema)

    # check that the update is successful
    res = client.table.get(name=test_table_name, schema="public")
    assert res.name == test_table_name
    assert res.table_schema == "public"
    assert len(res.columns) == 1
    assert res.constraints.primary_key.columns == ["number"]

    # move the object back to new_schema for testing
    table_object = BaseTable(schema=new_schema)
    res = client.table.update(name=test_table_name, table=table_object)
    res = client.table.get(name=test_table_name, schema=new_schema)
    assert res.name == test_table_name


def test_insert_records_invalid(client: ThanoSQL, new_schema: str):
    target_table = client.table.get(name=test_table_name, schema=new_schema)

    # inserting unsuitable records should result in IntegrityError
    with pytest.raises(ThanoSQLValueError):
        target_table.insert(records=[{"number": 1, "language": "Python"}])


def test_insert_get_records_success(client: ThanoSQL, new_schema: str):
    target_table = client.table.get(name=test_table_name, schema=new_schema)
    records = [{"number": i + 1} for i in range(5)]
    res = target_table.insert(records=records)

    # check if the records are successfully inserted
    res = target_table.get_records()
    assert {"records", "total"} == set(res.keys())
    assert res["total"] == len(records)
    assert res["records"] == records


def test_upload_table_invalid(client: ThanoSQL, new_schema: str):
    # only CSV or Excel accepted
    with pytest.raises(ThanoSQLValueError):
        client.table.upload(name=test_table_name, file="file_txt.txt")

    # schema should exist
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.upload(
            name=test_table_name, file="file_csv.csv", schema=fake.unique.pystr(10)
        )

    # if_exists should be valid
    with pytest.raises(ThanoSQLValueError):
        client.table.upload(
            name=test_table_name, file="file_csv.csv", if_exists="random"
        )

    # if the table exists but if_exists is fail, error should be raised
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.table.upload(
            name=test_table_name, file="file_csv.csv", schema=new_schema
        )

    # if the table exists and if_exists is append but data is not of appropriate format, error should be raised
    with pytest.raises(ThanoSQLValueError):
        client.table.upload(
            name=test_table_name,
            file="file_csv.csv",
            schema=new_schema,
            if_exists="append",
        )

    # if body does not match data, error should be raised
    with pytest.raises(ThanoSQLValueError):
        table_object = TableObject(
            columns=[BaseColumn(type="integer", name="number")],
            constraints=Constraints(primary_key=PrimaryKey(columns=["number"])),
        )
        client.table.upload(
            name=fake.unique.pystr(10), file="file_csv.csv", table=table_object
        )


def test_upload_table_csv(client: ThanoSQL, basic_table_name: str):
    client.table.upload(name=basic_table_name, file="file_csv.csv", if_exists="replace")

    # check that the table and records are indeed uploaded
    target_table = client.table.get(name=basic_table_name)
    assert len(target_table.columns) == 9

    # check get records with limit in the meantime
    limit = 5
    res = target_table.get_records(limit=limit)
    assert len(res["records"]) == limit


def test_upload_table_excel(client: ThanoSQL, new_schema: str):
    # using upload with a suitable body should work
    table_object = TableObject(
        columns=[
            BaseColumn(type="varchar", name="title"),
            BaseColumn(type="double precision", name="price"),
            BaseColumn(type="varchar", name="star_rating"),
        ]
    )

    client.table.upload(
        name=test_table_name_excel,
        file="file_excel.xlsx",
        schema=new_schema,
        table=table_object,
    )

    # check that the table and records are indeed uploaded
    target_table = client.table.get(name=test_table_name_excel, schema=new_schema)
    assert len(target_table.columns) == 3

    # check get records with limit in the meantime
    limit = 5
    res = target_table.get_records(limit=limit)
    assert len(res["records"]) == limit


def test_get_records_as_csv(client: ThanoSQL, new_schema: str):
    target_table = client.table.get(name=test_table_name_excel, schema=new_schema)
    target_table.get_records_as_csv()

    # check that the csv file is created
    csv_files = glob.glob("*[!file_csv].csv")
    assert len(csv_files) >= 1

    # check the contents of each csv file and remove it when done
    for csv_file in csv_files:
        with open(file=csv_file, mode="rb") as f:
            lines = f.readlines()
            assert len(lines) == 21

        try:
            os.remove(csv_file)
        except OSError:
            logging.info(f"{csv_file} is already removed")

    # we can technically check to see whether the contents of file_excel.xlsx and the recently-created csv are the same
    # using pandas or other library, but it adds more requirements and requires some time and resources
    # for now, we will just rely on whether a non-empty new csv file is created or not


def test_delete_table(client: ThanoSQL, new_schema: str):
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.delete(name=test_table_name, schema=fake.unique.pystr(10))

    # the other two tables will be automatically deleted when fixtures are destroyed
    # so we just need to "manually" delete these
    client.table.delete(name=test_table_name, schema=new_schema)
    client.table.delete(name=test_table_name_excel, schema=new_schema)

    # check that the table is indeed deleted
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.get(name=test_table_name, schema=new_schema)

    # trying to delete the table again should not be allowed
    with pytest.raises(ThanoSQLNotFoundError):
        client.table.delete(name=test_table_name, schema=new_schema)



from __future__ import annotations

import json
import os
from typing import Any, Optional, Union

import requests
from tqdm import tqdm

import thanosql._error as thanosql_error


class ThanoSQLBaseClient:
    def __init__(self, base_url: str, version: str, token: str) -> None:
        self.base_url: str = base_url.strip("/")
        self.version: str = version
        self.token: str = token

        self.url: str = f"{self.base_url}/api/{version}"

    def _create_auth_header(self) -> dict:
        return {"Authorization": f"Bearer {self.token}"}

    def _create_full_url(
        self,
        path: str = "",
        path_params: Optional[dict] = None,
        query_params: Optional[dict] = None,
    ) -> str:
        url = self.url + path

        if path_params:
            for param, value in path_params.items():
                url = url.replace(param, value)

        if query_params:
            query_params_list = []
            for param, value in query_params.items():
                query_params_list.append(f"{param}={value}")
            query_params_string = "&".join(query_params_list)
            url = f"{url}?{query_params_string}"

        return url

    def _request(
        self,
        method: str,
        path: str,
        path_params: Optional[dict] = None,
        query_params: Optional[dict] = None,
        payload: Optional[dict] = None,
        file: Optional[Union[str, os.PathLike]] = None,
        stream: bool = False,
    ) -> Any:
        full_url = self._create_full_url(
            path=path, path_params=path_params, query_params=query_params
        )

        headers = self._create_auth_header()
        headers["accept"] = "application/json"

        payload_json = {}

        try:
            if file:
                payload_json["files"] = {"file": (file, open(file, "rb"))}
                if payload:
                    payload_json["files"]["body"] = (
                        None,
                        json.dumps(payload),
                        "application/json",
                    )

            elif payload is not None:
                payload_json["json"] = payload

            request_func = getattr(requests, method.lower())
            response = request_func(
                url=full_url, headers=headers, stream=stream, **payload_json
            )

            response_json = {}
            if "application/json" in response.headers.get("Content-Type", ""):
                response_json = response.json()

            if not response.ok or "error" in response_json:
                code = response.status_code
                message = "An error had occurred. Please contact ThanoSQL team."

                if "error" in response_json:
                    code = response_json["error"].get("code", code)
                    message = response_json["error"].get("message", message)

                if code in {400, 405, 422}:
                    raise thanosql_error.ThanoSQLValueError(message=message)
                elif code in {401, 403}:
                    raise thanosql_error.ThanoSQLPermissionError(message=message)
                elif code == 404:
                    raise thanosql_error.ThanoSQLNotFoundError(message=message)
                elif code == 409:
                    raise thanosql_error.ThanoSQLAlreadyExistsError(message=message)
                elif code == 500:
                    raise thanosql_error.ThanoSQLInternalError(message=message)

            if stream:
                filename = response.headers.get(
                    "Content-Disposition", "filename=output.bin"
                ).split("filename=")[1]
                with open(filename, "wb") as handle:
                    for data in tqdm(response.iter_content()):
                        handle.write(data)
                return {"message": f"Successfully downloaded {filename}."}

            if response_json:
                return response_json

            return response
        except thanosql_error.ThanoSQLError as ex:
            raise ex
        except FileNotFoundError as ex:
            raise thanosql_error.ThanoSQLNotFoundError(message=str(ex))
        except PermissionError as ex:
            raise thanosql_error.ThanoSQLPermissionError(message=str(ex))
        except requests.exceptions.JSONDecodeError as ex:
            raise thanosql_error.ThanoSQLValueError(message=str(ex))
        except TypeError as ex:
            raise thanosql_error.ThanoSQLValueError(message=str(ex))
        except Exception as e:
            raise thanosql_error.ThanoSQLInternalError(message=str(e))



from __future__ import annotations

import os

from thanosql._base_client import ThanoSQLBaseClient
from thanosql.resources import (
    FileService,
    QueryService,
    SchemaService,
    TableService,
    ViewService,
)

THANOSQL_API_TOKEN: str = os.environ.get("THANOSQL_API_TOKEN", "")
THANOSQL_ENGINE_URL: str = os.environ.get("THANOSQL_ENGINE_URL", "")
THANOSQL_API_VERSION: str = "v1"


class ThanoSQL(ThanoSQLBaseClient):
    def __init__(
        self,
        engine_url: str = THANOSQL_ENGINE_URL,
        api_version: str = THANOSQL_API_VERSION,
        api_token: str = THANOSQL_API_TOKEN,
    ) -> None:
        super().__init__(base_url=engine_url, version=api_version, token=api_token)

        self.file: FileService = FileService(self)
        self.query: QueryService = QueryService(self)
        self.schema: SchemaService = SchemaService(self)
        self.table: TableService = TableService(self)
        self.view: ViewService = ViewService(self)



from typing import Optional


class ThanoSQLError(Exception):
    def __init__(
        self,
        message: Optional[str] = None,
        json_body: Optional[object] = None,
        code: Optional[str] = None,
    ):
        self.message: Optional[str] = message
        self.json_body: Optional[object] = json_body
        self.code: Optional[str] = code

    def __str__(self):
        msg = self.message or "<empty message>"
        return msg


class ThanoSQLAlreadyExistsError(ThanoSQLError):
    pass


class ThanoSQLConnectionError(ThanoSQLError):
    pass


class ThanoSQLInternalError(ThanoSQLError):
    pass


class ThanoSQLNotFoundError(ThanoSQLError):
    pass


class ThanoSQLPermissionError(ThanoSQLError):
    pass


class ThanoSQLValueError(ThanoSQLError):
    pass



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional, Union

from thanosql._service import ThanoSQLService

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class FileService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="file")

    def list(self, path: Union[str, os.PathLike]) -> dict:
        api_path = f"/{self.tag}/"
        query_params = self._create_input_dict(file_path=path)

        return self.client._request(
            method="get", path=api_path, query_params=query_params
        )

    def upload(
        self,
        path: Union[str, os.PathLike],
        db_commit: Optional[bool] = None,
        table: Optional[str] = None,
        column: Optional[str] = None,
        dir: Optional[str] = None,
    ) -> dict:
        api_path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            db_commit=db_commit, table_name=table, column_name=column, dir=dir
        )

        return self.client._request(
            method="post", path=api_path, query_params=query_params, file=path
        )

    def delete(
        self,
        path: Union[str, os.PathLike],
        db_commit: Optional[bool] = None,
        table: Optional[str] = None,
        column: Optional[str] = None,
    ) -> dict:
        api_path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            file_path=path, db_commit=db_commit, table_name=table, column_name=column
        )

        return self.client._request(
            method="delete", path=api_path, query_params=query_params
        )



from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from thanosql._error import ThanoSQLValueError
from thanosql._service import ThanoSQLService

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class BaseColumn(BaseModel):
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Column(BaseModel):
    id: Optional[int] = None
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Unique(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class PrimaryKey(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class ForeignKey(BaseModel):
    name: Optional[str] = None
    reference_schema: str = "public"
    reference_column: str
    reference_table: str
    column: str


class Constraints(BaseModel):
    unique: Optional[List[Unique]] = None
    primary_key: Optional[PrimaryKey] = None
    foreign_keys: Optional[List[ForeignKey]] = None


class BaseTable(BaseModel):
    name: Optional[str] = None
    table_schema: Optional[str] = Field(alias="schema", default=None)
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class TableObject(BaseModel):
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class TableService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table")

        self.template: TableTemplateService = TableTemplateService(client)

    def list(
        self,
        schema: Optional[str] = None,
        verbose: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Union[List[BaseTable], dict]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema, verbose=verbose, offset=offset, limit=limit
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "tables" in raw_response:
            tables_adapter = TypeAdapter(List[BaseTable])
            parsed_response = tables_adapter.validate_python(
                raw_response["tables"]
            )
            return parsed_response

        return raw_response

    def get(self, name: str, schema: Optional[str] = None) -> Union[Table, dict]:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "table" in raw_response:
            table_adapter = TypeAdapter(Table)
            parsed_response = table_adapter.validate_python(
                raw_response["table"]
            )
            parsed_response.service = self
            return parsed_response

        return raw_response

    def update(
        self, name: str, schema: Optional[str] = None, table: Optional[BaseTable] = None
    ) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        return self.client._request(
            method="put", path=path, query_params=query_params, payload=payload
        )

    def create(
        self,
        name: str,
        schema: Optional[str] = None,
        table: Optional[TableObject] = None,
    ) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        return self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

    def upload(
        self,
        name: str,
        file: Union[str, os.PathLike],
        schema: Optional[str] = None,
        table: Optional[TableObject] = None,
        if_exists: Optional[str] = None,
    ) -> dict:
        path = f"/{self.tag}/{name}/upload/"

        file_extension = Path(file).suffix.lower()
        if file_extension == ".csv":
            path = path + "csv"
        elif file_extension in [
            ".xls",
            ".xlsx",
            ".xlsm",
            ".xlsb",
            ".odf",
            ".ods",
            ".odt",
        ]:
            path = path + "excel"
        else:
            raise ThanoSQLValueError(
                "Invalid format: only CSV and Excel files possible."
            )

        query_params = self._create_input_dict(schema=schema, if_exists=if_exists)
        payload = self._create_input_dict(table=table)

        return self.client._request(
            method="post",
            path=path,
            query_params=query_params,
            payload=payload,
            file=file,
        )

    def delete(self, name: str, schema: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )


class TableTemplate(BaseModel):
    name: str
    table_template: TableObject
    version: Optional[str]
    compatibility: Optional[str]
    created_at: Optional[datetime]


class TableTemplateService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table_template")

    def list(
        self,
        search: Optional[str] = None,
        order_by: Optional[str] = None,
        latest: Optional[bool] = None,
    ) -> Union[List[TableTemplate], dict]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            search=search,
            order_by=order_by,
            latest=latest,
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "table_templates" in raw_response:
            table_templates_adapter = TypeAdapter(List[TableTemplate])
            parsed_response = table_templates_adapter.validate_python(
                raw_response["table_templates"]
            )
            return parsed_response

        return raw_response

    def get(self, name: str, version: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        if "table_templates" in raw_response:
            table_templates_adapter = TypeAdapter(List[TableTemplate])
            parsed_response = {}
            parsed_response[
                "table_templates"
            ] = table_templates_adapter.validate_python(
                raw_response["table_templates"]
            )
            parsed_response["versions"] = raw_response["versions"]
            return parsed_response

        return raw_response

    def create(
        self,
        name: str,
        table_template: TableObject,
        version: Optional[str] = None,
        compatibility: Optional[str] = None,
    ) -> dict:
        path = f"/{self.tag}/{name}"
        payload = self._create_input_dict(
            table_template=vars(table_template),
            version=version,
            compatibility=compatibility,
        )

        return self.client._request(method="post", path=path, payload=payload)

    def delete(self, name: str, version: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )


class Table(BaseTable):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    service: Optional[TableService] = None

    def get_records(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> dict:
        path = f"/{self.service.tag}/{self.name}/records"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            offset=offset,
            limit=limit,
        )

        return self.service.client._request(
            method="get",
            path=path,
            query_params=query_params,
        )

    def get_records_as_csv(
        self,
        timezone_offset: Optional[int] = None,
    ) -> None:
        path = f"/{self.service.tag}/{self.name}/records/csv"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            timezone_offset=timezone_offset,
        )

        self.service.client._request(
            method="get", path=path, query_params=query_params, stream=True
        )

    def insert(
        self,
        records: List[dict],
    ) -> dict:
        path = f"/{self.service.tag}/{self.name}/records"
        query_params = self.service._create_input_dict(schema=self.table_schema)

        return self.service.client._request(
            method="post", path=path, query_params=query_params, payload=records
        )



from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

pkg_vars = {}
with open("thanosql/_version.py") as f:
    exec(f.read(), pkg_vars)

setup(
    name="thanosql",
    version=pkg_vars["__version__"],
    description="ThanoSQL SDK for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/smartmind-team/thanosql-python",
    author="SmartMind",
    author_email="dev@smartmind.team",
    license="MIT",
    python_requires=">= 3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="smartmind thanosql sdk",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=["requests", "urllib3", "pydantic>=2.0", "tqdm"],
    extras_require={
        "dev": [
            "pytest",
            "faker",
        ],
        "magic": [
            "ipython",
            "pandas",
            "sqlalchemy",
            "numpy",
            "matplotlib",
            "websocket-client",
            "pglast",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)
